#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------------------------------------------------
# Author: Lalith Kumar Shiyam Sundar
#         Sebastian Gutschmayer
# Institution: Medical University of Vienna
# Research Group: Quantitative Imaging and Medical Physics (QIMP) Team
# Date: 09.02.2023
# Version: 2.0.0
#
# Description:
# This module handles image conversion for the moosez.
#
# Usage:
# The functions in this module can be imported and used in other modules within the moosez to perform image conversion.
#
# ----------------------------------------------------------------------------------------------------------------------

import contextlib
import io
import os
import re
import shutil
import unicodedata

import SimpleITK
import dicom2nifti
import pydicom
from rich.progress import Progress


def read_dicom_folder(folder_path: str) -> SimpleITK.Image:
    """
    Reads a folder of DICOM files and returns the image.

    :param folder_path: str
        The path to the folder containing the DICOM files.
    :type folder_path: str

    :return: SimpleITK.Image
        The image obtained from the DICOM files.
    :rtype: SimpleITK.Image
    """
    reader = SimpleITK.ImageSeriesReader()
    dicom_names = reader.GetGDCMSeriesFileNames(folder_path)
    reader.SetFileNames(dicom_names)

    dicom_image = reader.Execute()
    return dicom_image


def non_nifti_to_nifti(input_path: str, output_directory: str = None) -> None:
    """
    Converts any image format known to ITK to NIFTI

        :param input_path: The path to the directory or filename to convert to nii.gz.
        :type input_path: str
        
        :param output_directory: Optional. The output directory to write the image to. If not specified, the output image will be written to the same directory as the input image.
        :type output_directory: str
        
        :return: None
        :rtype: None
        
        :raises: FileNotFoundError if the input path does not exist.
        
        Usage:
        This function can be used to convert any image format known to ITK to NIFTI. If the input path is a directory, the function will convert all images in the directory to NIFTI format. If the input path is a file, the function will convert the file to NIFTI format. The output image will be written to the specified output directory, or to the same directory as the input image if no output directory is specified.
    """

    if not os.path.exists(input_path):
        print(f"Input path {input_path} does not exist.")
        return

    # Processing a directory
    if os.path.isdir(input_path):
        dicom_info = create_dicom_lookup(input_path)
        nifti_dir = dcm2niix(input_path)
        rename_nifti_files(nifti_dir, dicom_info)
        return

    # Processing a file
    if os.path.isfile(input_path):
        # Ignore hidden or already processed files
        _, filename = os.path.split(input_path)
        if filename.startswith('.') or filename.endswith(('.nii.gz', '.nii')):
            return
        else:
            output_image = SimpleITK.ReadImage(input_path)
            output_image_basename = f"{os.path.splitext(filename)[0]}.nii"

    if output_directory is None:
        output_directory = os.path.dirname(input_path)

    output_image_path = os.path.join(output_directory, output_image_basename)
    SimpleITK.WriteImage(output_image, output_image_path)


def standardize_to_nifti(parent_dir: str) -> None:
    """
    Converts all non-NIfTI images in a parent directory and its subdirectories to NIfTI format.

    :param parent_dir: The path to the parent directory containing the images to convert.
    :type parent_dir: str
    :return: None
    """
    # Get a list of all subdirectories in the parent directory
    subjects = os.listdir(parent_dir)
    subjects = [subject for subject in subjects if os.path.isdir(os.path.join(parent_dir, subject))]

    # Convert all non-NIfTI images in each subdirectory to NIfTI format
    with Progress() as progress:
        task = progress.add_task("[white] Processing subjects...", total=len(subjects))
        for subject in subjects:
            subject_path = os.path.join(parent_dir, subject)
            if os.path.isdir(subject_path):
                images = os.listdir(subject_path)
                for image in images:
                    image_path = os.path.join(subject_path, image)
                    path_is_valid = os.path.isdir(image_path) | os.path.isfile(image_path)
                    path_is_valid = path_is_valid & ("moosez" not in os.path.basename(image_path))
                    if path_is_valid:
                        non_nifti_to_nifti(image_path)
            else:
                continue
            progress.update(task, advance=1, description=f"[white] Processing {subject}...")


def dcm2niix(input_path: str) -> str:
    """
    Converts DICOM images into NIfTI images using dcm2niix.

    :param input_path: The path to the folder containing the DICOM files to convert.
    :type input_path: str
    :return: The path to the folder containing the converted NIfTI files.
    :rtype: str
    """
    output_dir = os.path.dirname(input_path)

    # Redirect standard output and standard error to discard output
    with contextlib.redirect_stdout(io.StringIO()), contextlib.redirect_stderr(io.StringIO()):
        dicom2nifti.convert_directory(input_path, output_dir, compression=False, reorient=True)

    return output_dir


def remove_accents(unicode_filename: str) -> str:
    """
    Removes accents and special characters from a Unicode filename.

    :param unicode_filename: The Unicode filename to clean.
    :type unicode_filename: str
    :return: The cleaned filename.
    :rtype: str
    """
    try:
        unicode_filename = str(unicode_filename).replace(" ", "_")
        cleaned_filename = unicodedata.normalize('NFKD', unicode_filename).encode('ASCII', 'ignore').decode('ASCII')
        cleaned_filename = re.sub(r'[^\w\s-]', '', cleaned_filename.strip().lower())
        cleaned_filename = re.sub(r'[-\s]+', '-', cleaned_filename)
        return cleaned_filename
    except:
        return unicode_filename


def is_dicom_file(filename: str) -> bool:
    """
    Checks if a file is a DICOM file.

    :param filename: The path to the file to check.
    :type filename: str
    :return: True if the file is a DICOM file, False otherwise.
    :rtype: bool
    """
    try:
        pydicom.dcmread(filename)
        return True
    except pydicom.errors.InvalidDicomError:
        return False


def create_dicom_lookup(dicom_dir: str) -> dict:
    """
    Create a lookup dictionary from DICOM files.

    :param dicom_dir: The directory where DICOM files are stored.
    :type dicom_dir: str
    :return: A dictionary where the key is the anticipated filename that dicom2nifti will produce and
             the value is the modality of the DICOM series.
    :rtype: dict
    """
    dicom_info = {}

    for filename in os.listdir(dicom_dir):
        full_path = os.path.join(dicom_dir, filename)
        if is_dicom_file(full_path):
            ds = pydicom.dcmread(full_path)

            series_number = ds.SeriesNumber if 'SeriesNumber' in ds else False
            series_description = ds.SeriesDescription if 'SeriesDescription' in ds else False
            sequence_name = ds.SequenceName if 'SequenceName' in ds else False
            protocol_name = ds.ProtocolName if 'ProtocolName' in ds else False
            series_instance_UID = ds.SeriesInstanceUID if 'SeriesInstanceUID' in ds else None

            modality = ds.Modality

            if series_number is not False:
                base_filename = remove_accents(series_number)
                if series_description is not False:
                    anticipated_filename = f"{base_filename}_{remove_accents(series_description)}.nii"
                elif sequence_name is not False:
                    anticipated_filename = f"{base_filename}_{remove_accents(sequence_name)}.nii"
                elif protocol_name is not False:
                    anticipated_filename = f"{base_filename}_{remove_accents(protocol_name)}.nii"
            else:
                anticipated_filename = f"{remove_accents(series_instance_UID)}.nii"

            dicom_info[anticipated_filename] = modality

    return dicom_info


def rename_nifti_files(nifti_dir: str, dicom_info: dict) -> None:
    """
    Rename NIfTI files based on a lookup dictionary.

    :param nifti_dir: The directory where NIfTI files are stored.
    :type nifti_dir: str
    :param dicom_info: A dictionary where the key is the anticipated filename that dicom2nifti will produce and
                       the value is the modality of the DICOM series.
    :type dicom_info: dict
    """
    for filename in os.listdir(nifti_dir):
        if filename.endswith('.nii'):
            modality = dicom_info.get(filename, '')
            if modality:
                new_filename = f"{modality}_{filename}"
                old_filepath = os.path.join(nifti_dir, filename)
                new_filepath = os.path.join(nifti_dir, new_filename)

                # Move and overwrite the file if it already exists
                shutil.move(old_filepath, new_filepath)
                del dicom_info[filename]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------------------------------------------------
# Author: Lalith Kumar Shiyam Sundar
# Institution: Medical University of Vienna
# Research Group: Quantitative Imaging and Medical Physics (QIMP) Team
# Date: 09.02.2023
# Version: 2.0.0
#
# Description:
# The main module of the mooseZ. It contains the main function that is executed when the mooseZ is run.
#
# Usage:
# The main function in this module is executed when the mooseZ is run.
#
# ----------------------------------------------------------------------------------------------------------------------

import argparse
import glob
import logging
import os
import time
from datetime import datetime

import SimpleITK
import colorama
import emoji
from halo import Halo
from moosez import constants
from moosez import display
from moosez import download
from moosez import file_utilities
from moosez import image_conversion
from moosez import image_processing
from moosez import input_validation
from moosez import predict
from moosez import resources
from moosez.image_processing import ImageResampler
from moosez.nnUNet_custom_trainer.utility import add_custom_trainers_to_local_nnunetv2
from moosez.resources import MODELS, AVAILABLE_MODELS


def main():
    logging.basicConfig(format='%(asctime)s %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s', level=logging.INFO,
                        filename=datetime.now().strftime('moosez-v.2.0.0.%H-%M-%d-%m-%Y.log'), filemode='w')
    colorama.init()

    # Argument parser
    parser = argparse.ArgumentParser(
        description=display.get_usage_message(),
        formatter_class=argparse.RawTextHelpFormatter,  # To retain the custom formatting
        add_help=False  # We'll add our own help option later
    )

    # Main directory containing subject folders
    parser.add_argument(
        "-d", "--main_directory",
        type=str,
        required=True,
        metavar="<MAIN_DIRECTORY>",
        help="Specify the main directory containing subject folders."
    )

    # Name of the model to use for segmentation
    model_help_text = "Choose the model for segmentation from the following:\n" + "\n".join(AVAILABLE_MODELS)
    parser.add_argument(
        "-m", "--model_name",
        type=str,
        choices=AVAILABLE_MODELS,
        required=True,
        metavar="<MODEL_NAME>",
        help=model_help_text
    )

    # Custom help option
    parser.add_argument(
        "-h", "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit."
    )

    args = parser.parse_args()

    parent_folder = os.path.abspath(args.main_directory)
    model_name = args.model_name

    display.logo()
    display.citation()

    logging.info('----------------------------------------------------------------------------------------------------')
    logging.info('                                     STARTING MOOSE-Z V.2.0.0                                       ')
    logging.info('----------------------------------------------------------------------------------------------------')

    # ----------------------------------
    # INPUT VALIDATION AND PREPARATION
    # ----------------------------------

    logging.info(' ')
    logging.info('- Main directory: ' + parent_folder)
    logging.info('- Model name: ' + model_name)
    logging.info(' ')
    print(' ')
    print(f'{constants.ANSI_VIOLET} {emoji.emojize(":memo:")} NOTE:{constants.ANSI_RESET}')
    print(' ')
    modalities = display.expectations(model_name)
    custom_trainer_status = add_custom_trainers_to_local_nnunetv2()
    logging.info('- Custom trainer: ' + custom_trainer_status)
    accelerator = resources.check_device()

    # ----------------------------------
    # DOWNLOADING THE MODEL
    # ----------------------------------

    print('')
    print(f'{constants.ANSI_VIOLET} {emoji.emojize(":globe_with_meridians:")} MODEL DOWNLOAD:{constants.ANSI_RESET}')
    print('')
    model_path = constants.NNUNET_RESULTS_FOLDER
    file_utilities.create_directory(model_path)
    download.model(model_name, model_path)

    # ----------------------------------
    # INPUT STANDARDIZATION
    # ----------------------------------

    print('')
    print(
        f'{constants.ANSI_VIOLET} {emoji.emojize(":magnifying_glass_tilted_left:")} STANDARDIZING INPUT DATA TO NIFTI:{constants.ANSI_RESET}')
    print('')
    logging.info(' ')
    logging.info(' STANDARDIZING INPUT DATA TO NIFTI:')
    logging.info(' ')
    image_conversion.standardize_to_nifti(parent_folder)
    print(f"{constants.ANSI_GREEN} Standardization complete.{constants.ANSI_RESET}")
    logging.info(" Standardization complete.")

    # --------------------------------------
    # CHECKING FOR MOOSE COMPLIANT SUBJECTS
    # --------------------------------------

    subjects = [os.path.join(parent_folder, d) for d in os.listdir(parent_folder) if
                os.path.isdir(os.path.join(parent_folder, d))]
    moose_compliant_subjects = input_validation.select_moose_compliant_subjects(subjects, modalities)

    num_subjects = len(moose_compliant_subjects)

    if num_subjects < 1:
        print(
            f'{constants.ANSI_RED} {emoji.emojize(":cross_mark:")} No moose compliant subject found to continue!{constants.ANSI_RESET} {emoji.emojize(":light_bulb:")} See: https://github.com/QIMP-Team/MOOSE#directory-structure-and-naming-conventions-for-moose-%EF%B8%8F')
        return

    # -------------------------------------------------
    # RUN PREDICTION ONLY FOR MOOSE COMPLIANT SUBJECTS
    # -------------------------------------------------

    print('')
    print(f'{constants.ANSI_VIOLET} {emoji.emojize(":crystal_ball:")} PREDICT:{constants.ANSI_RESET}')
    print('')
    logging.info(' ')
    logging.info(' PERFORMING PREDICTION:')
    logging.info(' ')

    spinner = Halo(text=' Initiating', spinner='dots')
    spinner.start()
    start_total_time = time.time()

    for i, subject in enumerate(moose_compliant_subjects):
        # SETTING UP DIRECTORY STRUCTURE
        spinner.text = f'[{i + 1}/{num_subjects}] Setting up directory structure for {os.path.basename(subject)}...'
        logging.info(' ')
        logging.info(f'{constants.ANSI_VIOLET} SETTING UP MOOSE-Z DIRECTORY:'
                     f'{constants.ANSI_RESET}')
        logging.info(' ')
        moose_dir, input_dirs, output_dir, stats_dir = file_utilities.moose_folder_structure(subject, model_name,
                                                                                             modalities)
        logging.info(f" MOOSE directory for subject {os.path.basename(subject)} at: {moose_dir}")

        # ORGANISE DATA ACCORDING TO MODALITY
        spinner.text = f'[{i + 1}/{num_subjects}] Organising data according to modality for {os.path.basename(subject)}...'
        file_utilities.organise_files_by_modality([subject], modalities, moose_dir)

        # PREPARE THE DATA FOR PREDICTION
        spinner.text = f'[{i + 1}/{num_subjects}] Preparing data for prediction for {os.path.basename(subject)}...'
        for input_dir in input_dirs:
            input_validation.make_nnunet_compatible(input_dir)
        logging.info(f" {constants.ANSI_GREEN}Data preparation complete using {model_name} for subject "
                     f"{os.path.basename(subject)}{constants.ANSI_RESET}")

        # RUN PREDICTION
        start_time = time.time()
        logging.info(' ')
        logging.info(' RUNNING PREDICTION:')
        logging.info(' ')
        spinner.text = f'[{i + 1}/{num_subjects}] Running prediction for {os.path.basename(subject)} using {model_name}...'

        for input_dir in input_dirs:
            predict.predict(model_name, input_dir, output_dir, accelerator)
        logging.info(f"Prediction complete using {model_name}.")

        end_time = time.time()
        elapsed_time = end_time - start_time
        spinner.text = f' {constants.ANSI_GREEN}[{i + 1}/{num_subjects}] Prediction done for {os.path.basename(subject)} using {model_name}!' \
                       f' | Elapsed time: {round(elapsed_time / 60, 1)} min{constants.ANSI_RESET}'
        time.sleep(3)
        logging.info(
            f' {constants.ANSI_GREEN}[{i + 1}/{num_subjects}] Prediction done for {os.path.basename(subject)} using {model_name}!' f' | Elapsed time: {round(elapsed_time / 60, 1)} min{constants.ANSI_RESET}')
        # ----------------------------------
        # EXTRACT PET ACTIVITY
        # ----------------------------------
        pet_file = file_utilities.find_pet_file(subject)
        if pet_file is not None:
            pet_image = SimpleITK.ReadImage(pet_file)
            spinner.text = f'[{i + 1}/{num_subjects}] Extracting PET activity for {os.path.basename(subject)}...'
            multilabel_file = glob.glob(os.path.join(output_dir, MODELS[model_name]["multilabel_prefix"] + '*nii*'))[0]
            multilabel_image = SimpleITK.ReadImage(multilabel_file)
            resampled_multilabel_image = ImageResampler.reslice_identity(reference_image=pet_image,
                                                                         moving_image=multilabel_image,
                                                                         is_label_image=True)
            out_csv = os.path.join(stats_dir, os.path.basename(subject) + '_pet_activity.csv')
            image_processing.get_intensity_statistics(pet_image, resampled_multilabel_image, model_name, out_csv)
            spinner.text = f'{constants.ANSI_GREEN} [{i + 1}/{num_subjects}] PET activity extracted for {os.path.basename(subject)}! ' \
                           f'{constants.ANSI_RESET}'
            time.sleep(3)

        # ----------------------------------
        # EXTRACT VOLUME STATISTICS
        # ----------------------------------
        spinner.text = f'[{i + 1}/{num_subjects}] Extracting CT volume statistics for {os.path.basename(subject)}...'
        multilabel_file = glob.glob(os.path.join(output_dir, MODELS[model_name]["multilabel_prefix"] + '*nii*'))[0]
        multilabel_image = SimpleITK.ReadImage(multilabel_file)
        out_csv = os.path.join(stats_dir, os.path.basename(subject) + '_ct_volume.csv')
        image_processing.get_shape_statistics(multilabel_image, model_name, out_csv)
        spinner.text = f'{constants.ANSI_GREEN} [{i + 1}/{num_subjects}] CT volume extracted for {os.path.basename(subject)}! ' \
                       f'{constants.ANSI_RESET}'
        time.sleep(1)

    end_total_time = time.time()
    total_elapsed_time = (end_total_time - start_total_time) / 60
    time_per_dataset = total_elapsed_time / len(moose_compliant_subjects)

    spinner.succeed(f'{constants.ANSI_GREEN} All predictions done! | Total elapsed time for '
                    f'{len(moose_compliant_subjects)} datasets: {round(total_elapsed_time, 1)} min'
                    f' | Time per dataset: {round(time_per_dataset, 2)} min {constants.ANSI_RESET}')


def moose(model_name: str, input_dir: str, output_dir: str, accelerator: str) -> None:
    """
    Execute the MOOSE 2.0 image segmentation process.

    This function carries out the following steps:
    1. Sets the path for model results.
    2. Creates the required directory for the model.
    3. Downloads the model based on the provided `model_name`.
    4. Validates and prepares the input directory to be compatible with nnUNet.
    5. Executes the prediction process.

    :param model_name: The name of the model to be used for predictions. This model will be downloaded and used 
                       for the image segmentation process.
    :type model_name: str

    :param input_dir: Path to the directory containing the images (in nifti, either .nii or .nii.gz) to be processed.
    :type input_dir: str

    :param output_dir: Path to the directory where the segmented output will be saved.
    :type output_dir: str

    :param accelerator: Specifies the type of accelerator to be used. Common values include "cpu" and "cuda" for 
                        GPU acceleration.
    :type accelerator: str

    :return: None
    :rtype: None

    :Example:

    >>> moose('clin_ct_organs', '/path/to/input/images', '/path/to/save/output', 'cuda')

    """
    model_path = constants.NNUNET_RESULTS_FOLDER
    file_utilities.create_directory(model_path)
    download.model(model_name, model_path)
    custom_trainer_status = add_custom_trainers_to_local_nnunetv2()
    logging.info('- Custom trainer: ' + custom_trainer_status)
    input_validation.make_nnunet_compatible(input_dir)
    predict.predict(model_name, input_dir, output_dir, accelerator)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------------------------------------------------
# Author: Lalith Kumar Shiyam Sundar
# Institution: Medical University of Vienna
# Research Group: Quantitative Imaging and Medical Physics (QIMP) Team
# Date: 09.02.2023
# Version: 2.0.0
#
# Description:
# This module performs input validation for the moosez. It verifies that the inputs provided by the user are valid
# and meets the required specifications.
#
# Usage:
# The functions in this module can be imported and used in other modules within the moosez to perform input validation.
#
# ----------------------------------------------------------------------------------------------------------------------

import logging
import multiprocessing
import os
from typing import List

import nibabel as nib
from moosez import constants


def check_directory_exists(directory_path: str) -> None:
    """
    Checks if the specified directory exists.

    :param directory_path: The path to the directory.
    :type directory_path: str
    :raises: Exception if the directory does not exist.
    """
    if not os.path.isdir(directory_path):
        raise Exception(f"Error: The directory '{directory_path}' does not exist.")


def select_moose_compliant_subjects(subject_paths: List[str], modality_tags: List[str]) -> List[str]:
    """
    Selects the subjects that have the files that have names that are compliant with the moosez.

    :param subject_paths: The path to the list of subjects that are present in the parent directory.
    :type subject_paths: List[str]
    :param modality_tags: The list of appropriate modality prefixes that should be attached to the files for them to be moose compliant.
    :type modality_tags: List[str]
    :return: The list of subject paths that are moose compliant.
    :rtype: List[str]
    """
    # go through each subject in the parent directory
    moose_compliant_subjects = []
    for subject_path in subject_paths:
        # go through each subject and see if the files have the appropriate modality prefixes

        files = [file for file in os.listdir(subject_path) if file.endswith('.nii') or file.endswith('.nii.gz')]
        prefixes = [file.startswith(tag) for tag in modality_tags for file in files]
        if sum(prefixes) == len(modality_tags):
            moose_compliant_subjects.append(subject_path)
    print(f"{constants.ANSI_ORANGE} Number of moose compliant subjects: {len(moose_compliant_subjects)} out of "
          f"{len(subject_paths)} {constants.ANSI_RESET}")
    logging.info(f" Number of moose compliant subjects: {len(moose_compliant_subjects)} out of "
                 f"{len(subject_paths)}")

    return moose_compliant_subjects


def check_file_for_nnunet_compatibility(filename: str, input_dir: str) -> None:
    """
    Checks if the file is nnUNet compatible. If not, compresses the file and renames it to include the required tag.

    :param filename: The name of the file to check.
    :type filename: str
    :param input_dir: The path to the directory containing the file.
    :type input_dir: str
    :return: None
    :rtype: None
    """
    if filename.endswith('.nii.gz') and not filename.endswith('_0000.nii.gz'):
        # Rename the file to include the required tag
        new_filename = filename.replace('.nii.gz', '_0000.nii.gz')
        os.rename(os.path.join(input_dir, filename), os.path.join(input_dir, new_filename))
    elif not filename.endswith('.gz'):
        # Compress the file if it is not already compressed
        img = nib.load(os.path.join(input_dir, filename))
        new_filename = filename + '.gz'
        nib.save(img, os.path.join(input_dir, new_filename))
        os.remove(os.path.join(input_dir, filename))
        if not new_filename.endswith('_0000.nii.gz'):
            # Rename the file to include the required tag
            new_filename_with_zeroes = new_filename.replace('.nii.gz', '_0000.nii.gz')
            os.rename(os.path.join(input_dir, new_filename), os.path.join(input_dir, new_filename_with_zeroes))


def make_nnunet_compatible(input_dir: str) -> None:
    """
    Checks the files in the specified directory to ensure they comply with the nnUNet requirements. If a file does not
    comply, it is compressed (if it is not already) and renamed to include the required tag.

    :param input_dir: The path to the directory containing the files to check.
    :type input_dir: str
    :return: None
    :rtype: None
    """
    with multiprocessing.Pool() as pool:
        # Map the check_file() function to each file in the directory
        pool.starmap(check_file_for_nnunet_compatibility, [(filename, input_dir) for filename in os.listdir(input_dir)])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------------------------------------------------
# Author: Lalith Kumar Shiyam Sundar
# Institution: Medical University of Vienna
# Research Group: Quantitative Imaging and Medical Physics (QIMP) Team
# Date: 13.02.2023
# Version: 2.0.0
#
# Description:
# This module contains the necessary functions for prediction using the moosez.
#
# Usage:
# The functions in this module can be imported and used in other modules within the moosez to perform prediction.
#
# ----------------------------------------------------------------------------------------------------------------------

import glob
import os
import shutil
import subprocess
from typing import Tuple, List, Any

import nibabel as nib
import numpy as np
from halo import Halo
from moosez import constants
from moosez import file_utilities
from moosez import image_processing
from moosez.image_processing import ImageResampler
from moosez.image_processing import NiftiPreprocessor
from moosez.resources import MODELS, map_model_name_to_task_number
from mpire import WorkerPool


def predict(model_name: str, input_dir: str, output_dir: str, accelerator: str) -> None:
    """
    Runs the prediction using nnunet_predict.

    :param model_name: The name of the model.
    :type model_name: str
    :param input_dir: The input directory.
    :type input_dir: str
    :param output_dir: The output directory.
    :type output_dir: str
    :param accelerator: The accelerator to use.
    :type accelerator: str
    :return: None
    :rtype: None
    """
    task_number = map_model_name_to_task_number(model_name)
    # set the environment variables
    os.environ["nnUNet_results"] = constants.NNUNET_RESULTS_FOLDER

    # Preprocess the image
    temp_input_dir, resampled_image, moose_image_object = preprocess(input_dir, model_name)
    resampled_image_shape = resampled_image.shape
    resampled_image_affine = resampled_image.affine

    # choose the appropriate trainer for the model
    trainer = MODELS[model_name]["trainer"]

    # Construct the command
    command = f'nnUNetv2_predict -i {temp_input_dir} -o {output_dir} -d {task_number} -c 3d_fullres' \
              f' -f all -tr {trainer} --disable_tta -device {accelerator}'

    # Run the command
    subprocess.run(command, shell=True, stdout=subprocess.DEVNULL, env=os.environ, stderr=subprocess.DEVNULL)

    original_image_files = file_utilities.get_files(input_dir, '.nii.gz')

    # Postprocess the label
    # if temp_input_dir has more than one nifti file, run logic below

    if len(file_utilities.get_files(output_dir, '.nii.gz')) > 1:
        merge_image_parts(output_dir, resampled_image_shape, resampled_image_affine)
    postprocess(original_image_files[0], output_dir, model_name)

    shutil.rmtree(temp_input_dir)


def preprocess(original_image_directory: str, model_name: str) -> Tuple[str, nib.Nifti1Image, Any]:
    """
    Preprocesses the original images.

    :param original_image_directory: The directory containing the original images.
    :type original_image_directory: str
    :param model_name: The name of the model.
    :type model_name: str
    :return: A tuple containing the path to the temp folder, the resampled image, and the moose_image_object.
    :rtype: Tuple[str, nib.Nifti1Image, Any]
    """
    # create the temp directory
    temp_folder = os.path.join(original_image_directory, constants.TEMP_FOLDER)
    os.makedirs(temp_folder, exist_ok=True)
    original_image_files = file_utilities.get_files(original_image_directory, '.nii.gz')
    org_image = nib.load(original_image_files[0])
    moose_image_object = NiftiPreprocessor(org_image)

    # choose the target spacing for the model to enable preprocessing
    desired_spacing = MODELS[model_name]["voxel_spacing"]

    resampled_image = ImageResampler.resample_image(moose_img_object=moose_image_object,
                                                    interpolation=constants.INTERPOLATION,
                                                    desired_spacing=desired_spacing)
    # if model name has body in it, run logic below
    if "body" in model_name:
        image_processing.write_image(resampled_image, os.path.join(temp_folder, constants.RESAMPLED_IMAGE_FILE_NAME),
                                     False, False)
    else:
        image_processing.write_image(resampled_image, os.path.join(temp_folder, constants.RESAMPLED_IMAGE_FILE_NAME),
                                     moose_image_object.is_large, False)
    return temp_folder, resampled_image, moose_image_object


def postprocess(original_image: str, output_dir: str, model_name: str) -> None:
    """
    Postprocesses the predicted images.

    :param original_image: The path to the original image.
    :type original_image: str
    :param output_dir: The output directory containing the label image.
    :type output_dir: str
    :param model_name: The name of the model.
    :type model_name: str
    :return: None
    :rtype: None
    """
    # [1] Resample the predicted image to the original image's voxel spacing
    predicted_image = file_utilities.get_files(output_dir, '.nii.gz')[0]
    original_header = nib.load(original_image).header
    native_spacing = original_header.get_zooms()
    native_size = original_header.get_data_shape()
    resampled_prediction = ImageResampler.resample_segmentations(input_image_path=predicted_image,
                                                                 desired_spacing=native_spacing,
                                                                 desired_size=native_size)
    multilabel_image = os.path.join(output_dir, MODELS[model_name]["multilabel_prefix"] +
                                    os.path.basename(original_image))
    # if model_name has body in it, run logic below
    if "body" in model_name:
        resampled_prediction_data = resampled_prediction.get_fdata()
        resampled_prediction_new = nib.Nifti1Image(resampled_prediction_data, resampled_prediction.affine,
                                                   resampled_prediction.header)
        image_processing.write_image(resampled_prediction_new, multilabel_image, False, True)
    else:
        image_processing.write_image(resampled_prediction, multilabel_image, False, True)
    os.remove(predicted_image)


def count_output_files(output_dir: str) -> int:
    """
    Counts the number of files in the specified output directory.

    :param output_dir: The path to the output directory.
    :type output_dir: str
    :return: The number of files in the output directory.
    :rtype: int
    """
    return len([name for name in os.listdir(output_dir) if
                os.path.isfile(os.path.join(output_dir, name)) and name.endswith('.nii.gz')])


def monitor_output_directory(output_dir: str, total_files: int, spinner: Halo) -> None:
    """
    Continuously monitors the specified output directory for new files and updates the progress bar accordingly.

    :param output_dir: The path to the output directory.
    :type output_dir: str
    :param total_files: The total number of files that are expected to be generated in the output directory.
    :type total_files: int
    :param spinner: The spinner that displays the progress of the segmentation process.
    :type spinner: Halo
    :return: None
    :rtype: None
    """
    files_processed = count_output_files(output_dir)
    while files_processed < total_files:
        new_files_processed = count_output_files(output_dir)
        if new_files_processed > files_processed:
            spinner.text = f'Processed {new_files_processed} of {total_files} files'
            spinner.spinner = 'dots'
        files_processed = new_files_processed


def split_and_save(shared_image_data: Tuple[np.ndarray, np.ndarray], z_index: List[Tuple[int, int]],
                   image_chunk_path: str) -> None:
    """
    Split the image and save each part.

    :param shared_image_data: A tuple containing the voxel data of the image and the affine transformation associated with the data.
    :type shared_image_data: Tuple[np.ndarray, np.ndarray]
    :param z_index: A list of tuples containing start and end indices for z-axis split.
    :type z_index: List[Tuple[int, int]]
    :param image_chunk_path: The path to save the image part.
    :type image_chunk_path: str
    :return: None
    :rtype: None
    """
    image_data, image_affine = shared_image_data
    image_part = nib.Nifti1Image(image_data[:, :, z_index[0]:z_index[1]], image_affine)
    nib.save(image_part, image_chunk_path)


def handle_large_image(image: nib.Nifti1Image, save_dir: str) -> List[str]:
    """
    Split a large image into parts and save them.

    :param image: The NIBABEL image.
    :type image: nib.Nifti1Image
    :param save_dir: Directory to save the image parts.
    :type save_dir: str
    :return: List of paths of the original or split image parts.
    :rtype: List[str]
    """

    image_shape = image.shape
    image_data = image.get_fdata()
    image_affine = image.affine

    if np.prod(image_shape) > constants.MATRIX_THRESHOLD and image_shape[2] > constants.Z_AXIS_THRESHOLD:

        # Calculate indices for z-axis split
        z_part = image_shape[2] // 3
        z_indices = [(0, z_part + constants.MARGIN_PADDING),
                     (z_part + 1 - constants.MARGIN_PADDING, z_part * 2 + constants.MARGIN_PADDING),
                     (z_part * 2 + 1 - constants.MARGIN_PADDING, None)]
        filenames = ["subpart01_0000.nii.gz", "subpart02_0000.nii.gz", "subpart03_0000.nii.gz"]
        resampled_chunks_paths = [os.path.join(save_dir, filename) for filename in filenames]

        chunk_data = []
        for z_index, resampled_chunk_path in zip(z_indices, resampled_chunks_paths):
            chunk_data.append({"z_index": z_index,
                               "image_chunk_path": resampled_chunk_path})

        shared_objects = (image_data, image_affine)

        with WorkerPool(n_jobs=3, shared_objects=shared_objects) as pool:
            pool.map(split_and_save, chunk_data)

        return resampled_chunks_paths

    else:
        resampled_image_path = os.path.join(save_dir, constants.RESAMPLED_IMAGE_FILE_NAME)
        nib.save(image, resampled_image_path)

        return [resampled_image_path]


def merge_image_parts(save_dir: str, original_image_shape: Tuple[int, int, int],
                      original_image_affine: np.ndarray) -> str:
    """
    Combine the split image parts back into a single image.

    :param save_dir: Directory where the image parts are saved.
    :type save_dir: str
    :param original_image_shape: The shape of the original image.
    :type original_image_shape: Tuple[int, int, int]
    :param original_image_affine: The affine transformation of the original image.
    :type original_image_affine: np.ndarray
    :return: Path to the merged image.
    :rtype: str
    """
    # Create an empty array with the original image's shape
    merged_image_data = np.zeros(original_image_shape, dtype=np.uint8)

    # Calculate the split index along the z-axis
    z_split_index = original_image_shape[2] // 3

    # Load each part, extract its data, and place it in the correct position in the merged image

    predicted_chunk_filenames = [s.replace("_0000", "") for s in constants.CHUNK_FILENAMES]

    merged_image_data[:, :, :z_split_index] = nib.load(
        os.path.join(save_dir, predicted_chunk_filenames[0])).get_fdata()[:,
                                              :, :-constants.MARGIN_PADDING]
    merged_image_data[:, :, z_split_index:z_split_index * 2] = nib.load(
        os.path.join(save_dir, predicted_chunk_filenames[1])).get_fdata()[:, :,
                                                               constants.MARGIN_PADDING - 1:-constants.MARGIN_PADDING]
    merged_image_data[:, :, z_split_index * 2:] = nib.load(
        os.path.join(save_dir, predicted_chunk_filenames[2])).get_fdata()[
                                                  :, :, constants.MARGIN_PADDING - 1:]

    # Create a new Nifti1Image with the merged data and the original image's affine transformation
    merged_image = nib.Nifti1Image(merged_image_data, original_image_affine)

    # remove the split image parts
    files_to_remove = glob.glob(os.path.join(save_dir, constants.CHUNK_PREFIX + "*"))
    for file in files_to_remove:
        os.remove(file)

    # write the merged image to disk
    merged_image_path = os.path.join(save_dir, constants.RESAMPLED_IMAGE_FILE_NAME)
    nib.save(merged_image, merged_image_path)

    return merged_image_path



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------------------------------------------------
# Author: Lalith Kumar Shiyam Sundar
# Institution: Medical University of Vienna
# Research Group: Quantitative Imaging and Medical Physics (QIMP) Team
# Date: 09.02.2023
# Version: 2.0.0
#
# Description:
# This module shows predefined display messages for the moosez.
#
# Usage:
# The functions in this module can be imported and used in other modules within the moosez to show predefined display
# messages.
#
# ----------------------------------------------------------------------------------------------------------------------

import logging

import emoji
import pyfiglet
from moosez import constants
from moosez import resources


def get_usage_message():
    usage_message = """
    Usage:
      moosez -d <MAIN_DIRECTORY> -m <MODEL_NAME>
    Example:  
      moosez -d /Documents/Data_to_moose/ -m clin_ct_organs

    Description:
      MOOSE (Multi-organ objective segmentation) - A data-centric AI solution that
      generates multilabel organ segmentations for systemic TB whole-person research.
    """
    return usage_message


def logo():
    """
    Display MOOSE logo

    This function displays the MOOSE logo using the pyfiglet library and ANSI color codes.

    :return: None
    """
    print(' ')
    logo_color_code = constants.ANSI_VIOLET
    slogan_color_code = constants.ANSI_VIOLET
    result = logo_color_code + pyfiglet.figlet_format(" MOOSE 2.0", font="smslant").rstrip() + "\033[0m"
    text = slogan_color_code + " A part of the ENHANCE community. Join us at www.enhance.pet to build the future of " \
                               "PET imaging together." + "\033[0m"
    print(result)
    print(text)
    print(' ')


def citation():
    """
    Display manuscript citation

    This function displays the manuscript citation for the MOOSE project.

    :return: None
    """
    print(f'{constants.ANSI_VIOLET} {emoji.emojize(":scroll:")} CITATION:{constants.ANSI_RESET}')
    print(" ")
    print(
        " Shiyam Sundar LK, Yu J, Muzik O, et al. Fully-automated, semantic segmentation of whole-body 18F-FDG PET/CT "
        "images based on data-centric artificial intelligence. J Nucl Med. June 2022.")
    print(" Copyright 2022, Quantitative Imaging and Medical Physics Team, Medical University of Vienna")


def expectations(model_name: str) -> list:
    """
    Display expected modality for the model.

    This function displays the expected modality for the given model name. It also checks for a special case where
    'FDG-PET-CT' should be split into 'FDG-PET' and 'CT'.

    :param model_name: The name of the model.
    :type model_name: str
    :return: A list of modalities.
    :rtype: list
    """
    model_info = resources.expected_modality(model_name)
    modality = model_info['Modality']

    # check for special case where 'FDG-PET-CT' should be split into 'FDG-PET' and 'CT'
    if modality == 'FDG-PET-CT':
        modalities = ['FDG-PET', 'CT']
    else:
        modalities = [modality]
    expected_prefix = [m.replace('-', '_') + "_" for m in modalities]

    print(
        f" Imaging: {model_info['Imaging']} |"
        f" Modality: {modality} | "
        f"Tissue of interest: {model_info['Tissue of interest']}")
    print(
        f" Required modalities: {modalities} | "
        f" No. of modalities: {len(modalities)}"
        f" | Required prefix for non-DICOM files: {expected_prefix}")
    logging.info(f" Required modalities: {modalities} |  No. of modalities: {len(modalities)} "
                 f"| Required prefix for non-DICOM files: {expected_prefix} ")
    print(
        f"{constants.ANSI_ORANGE} Warning: Subjects which don't have the required modalities [check file prefix] "
        f"will be skipped. {constants.ANSI_RESET}")
    warning_message = " Skipping subjects without the required modalities (check file prefix).\n" \
                      " These subjects will be excluded from analysis and their data will not be used."
    logging.warning(warning_message)

    return modalities



import os
import platform
import site

from moosez import file_utilities


def add_custom_trainers_to_local_nnunetv2():
    # Locate the site-packages directory
    site_packages = site.getsitepackages()[0]

    if platform.system() == 'Windows':
        site_packages = os.path.join(site_packages, 'Lib', 'site-packages')

    source_file_path = os.path.join(site_packages, 'moosez', 'nnUNet_custom_trainer', 'MOOSE_custom_trainers.py')
    target_file_path = os.path.join(site_packages, 'nnunetv2', 'training', 'nnUNetTrainer', 'variants',
                                    'MOOSE_custom_trainers.py')

    # Check if the file exists
    if not os.path.exists(source_file_path):
        return f'Could not find custom trainer file: {source_file_path}'

    # Check if the file exists
    if os.path.exists(target_file_path):
        return f'Custom trainer already installed: {target_file_path}.'

    file_utilities.copy_file(source_file_path, target_file_path)

    return f"Custom trainer added to {target_file_path}"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
This module contains the constants that are used in the moosez.

.. moduleauthor:: Lalith Kumar Shiyam Sundar <lalith.shiyamsundar@meduniwien.ac.at>
.. versionadded:: 2.0.0
"""

import os

from moosez import file_utilities

project_root = file_utilities.get_virtual_env_root()

NNUNET_RESULTS_FOLDER = os.path.join(project_root, 'models', 'nnunet_trained_models')
MOOSEZ_MODEL_FOLDER = os.path.join(NNUNET_RESULTS_FOLDER, 'nnUNet', '3d_fullres')
ALLOWED_MODALITIES = ['CT', 'PT', 'MR']
TEMP_FOLDER = 'temp'

# COLOR CODES
ANSI_ORANGE = '\033[38;5;208m'
ANSI_GREEN = '\033[38;5;40m'
ANSI_VIOLET = '\033[38;5;141m'
ANSI_RED = '\033[38;5;196m'
ANSI_RESET = '\033[0m'

# SUPPORTED TRACERS (limited patch)

TRACER_FDG = 'FDG'

# FOLDER NAMES

SEGMENTATIONS_FOLDER = 'segmentations'
STATS_FOLDER = 'stats'

# PREPROCESSING PARAMETERS

MATRIX_THRESHOLD = 200 * 200 * 600
Z_AXIS_THRESHOLD = 200
MARGIN_PADDING = 20
INTERPOLATION = 'bspline'
CHUNK_THRESHOLD = 200

# POSTPROCESSING PARAMETERS

TUMOR_LABEL = 12

# FILE NAMES

RESAMPLED_IMAGE_FILE_NAME = 'resampled_image_0000.nii.gz'
RESAMPLED_MASK_FILE_NAME = 'resampled_mask.nii.gz'
CHUNK_FILENAMES = ["chunk01_0000.nii.gz", "chunk02_0000.nii.gz", "chunk03_0000.nii.gz"]
CHUNK_PREFIX = 'chunk'

# DISPLAY PARAMETERS

MIP_ROTATION_STEP = 40
DISPLAY_VOXEL_SPACING = (3, 3, 3)
FRAME_DURATION = 0.4

# ORGAN INDICES

ORGAN_INDICES = {
    "clin_ct_organs": {
        1: "spleen",
        2: "kidney_right",
        3: "kidney_left",
        4: "gallbladder",
        5: "liver",
        6: "stomach",
        7: "aorta",
        8: "inferior_vena_cava",
        9: "portal_vein_and_splenic_vein",
        10: "pancreas",
        11: "adrenal_gland_right",
        12: "adrenal_gland_left",
        13: "lung_upper_lobe_left",
        14: "lung_lower_lobe_left",
        15: "lung_upper_lobe_right",
        16: "lung_middle_lobe_right",
        17: "lung_lower_lobe_right"
    },
    "clin_ct_lungs": {
        1: "lung_upper_lobe_left",
        2: "lung_lower_lobe_left",
        3: "lung_upper_lobe_right",
        4: "lung_middle_lobe_right",
        5: "lung_lower_lobe_right"
    },
    "clin_ct_body": {
        1: "Legs",
        2: "Body",
        3: "Head",
        4: "Arms"
    },
    "clin_pt_fdg_tumor": {
        1: "tumor"
    },
    "preclin_mr_all": {
        1: "Brain",
        2: "Liver",
        3: "Intestines",
        4: "Pancreas",
        5: "Thyroid",
        6: "Spleen",
        7: "Bladder",
        8: "OuterKidney",
        9: "InnerKidney",
        10: "HeartInside",
        11: "HeartOutside",
        12: "WAT Subcutaneous",
        13: "WAT Visceral",
        14: "BAT",
        15: "Muscle TF",
        16: "Muscle TB",
        17: "Muscle BB",
        18: "Muscle BF",
        19: "Aorta",
        20: "Lung",
        21: "Stomach",
    },
    "clin_ct_ribs": {
        1: "rib_left_1",
        2: "rib_left_2",
        3: "rib_left_3",
        4: "rib_left_4",
        5: "rib_left_5",
        6: "rib_left_6",
        7: "rib_left_7",
        8: "rib_left_8",
        9: "rib_left_9",
        10: "rib_left_10",
        11: "rib_left_11",
        12: "rib_left_12",
        13: "rib_right_1",
        14: "rib_right_2",
        15: "rib_right_3",
        16: "rib_right_4",
        17: "rib_right_5",
        18: "rib_right_6",
        19: "rib_right_7",
        20: "rib_right_8",
        21: "rib_right_9",
        22: "rib_right_10",
        23: "rib_right_11",
        24: "rib_right_12",
        25: "sternum"
    },
    "clin_ct_muscles": {
        1: "gluteus_maximus_left",
        2: "gluteus_maximus_right",
        3: "gluteus_medius_left",
        4: "gluteus_medius_right",
        5: "gluteus_minimus_left",
        6: "gluteus_minimus_right",
        7: "autochthon_left",
        8: "autochthon_right",
        9: "iliopsoas_left",
        10: "iliopsoas_right"
    },
    "clin_ct_peripheral_bones": {
        1: "carpal_left",
        2: "carpal_right",
        3: "clavicle_left",
        4: "clavicle_right",
        5: "femur_left",
        6: "femur_right",
        7: "fibula_left",
        8: "fibula_right",
        9: "humerus_left",
        10: "humerus_right",
        11: "metacarpal_left",
        12: "metacarpal_right",
        13: "metatarsal_left",
        14: "metatarsal_right",
        15: "patella_left",
        16: "patella_right",
        17: "fingers_left",
        18: "fingers_right",
        19: "radius_left",
        20: "radius_right",
        21: "scapula_left",
        22: "scapula_right",
        23: "skull",
        24: "tarsal_left",
        25: "tarsal_right",
        26: "tibia_left",
        27: "tibia_right",
        28: "toes_left",
        29: "toes_right",
        30: "ulna_left",
        31: "ulna_right",
        32: "thyroid_left",
        33: "thyroid_right",
        34: "bladder"
    },
    "clin_ct_fat": {
        1: "spinal_chord",
        2: "skeletal_muscle",
        3: "subcutaneous_fat",
        4: "visceral_fat",
        5: "thoracic_fat",
        6: "eyes",
        7: "testicles",
        8: "prostate"
    },
    "clin_ct_vertebrae": {
        1: "vertebra_C1",
        2: "vertebra_C2",
        3: "vertebra_C3",
        4: "vertebra_C4",
        5: "vertebra_C5",
        6: "vertebra_C6",
        7: "vertebra_C7",
        8: "vertebra_T1",
        9: "vertebra_T2",
        10: "vertebra_T3",
        11: "vertebra_T4",
        12: "vertebra_T5",
        13: "vertebra_T6",
        14: "vertebra_T7",
        15: "vertebra_T8",
        16: "vertebra_T9",
        17: "vertebra_T10",
        18: "vertebra_T11",
        19: "vertebra_T12",
        20: "vertebra_L1",
        21: "vertebra_L2",
        22: "vertebra_L3",
        23: "vertebra_L4",
        24: "vertebra_L5",
        25: "vertebra_S1",
        26: "hip_left",
        27: "hip_right",
        28: "sacrum"
    },
    "clin_ct_cardiac": {
        1: "heart_myocardium",
        2: "heart_atrium_left",
        3: "heart_ventricle_left",
        4: "heart_atrium_right",
        5: "heart_ventricle_right",
        6: "pulmonary_artery",
        7: "iliac_artery_left",
        8: "iliac_artery_right",
        9: "iliac_vena_left",
        10: "iliac_vena_right"
    },
    "clin_ct_digestive": {
        1: "esophagus",
        2: "trachea",
        3: "small_bowel",
        4: "duodenum",
        5: "colon",
        6: "urinary_bladder",
        7: "face"
    },
    "preclin_ct_legs": {
        1: "right_leg_muscle",
        2: "left_leg_muscle"
    },
    "clin_ct_all_bones_v1": {
        1: "carpal",
        2: "clavicle",
        3: "femur",
        4: "fibula",
        5: "humerus",
        6: "metacarpal",
        7: "metatarsal",
        8: "patella",
        9: "pelvis",
        10: "fingers",
        11: "radius",
        12: "ribcage",
        13: "scapula",
        14: "skull",
        15: "spine",
        16: "sternum",
        17: "tarsal",
        18: "tibia",
        19: "toes",
        20: "ulna"
    },
    "clin_ct_PUMA": {
        0: "background",
        1: "Spleen",
        2: "Kidneys",
        3: "Gallbladder",
        4: "Liver",
        5: "Stomach",
        6: "Pancreas",
        7: "Adrenal Glands",
        8: "Lungs",
        9: "Heart",
        10: "Vessels",
        11: "Esophagus",
        12: "Trachea",
        13: "Small_bowel",
        14: "Duodenum",
        15: "Colon",
        16: "Brain",
        17: "Skeleton",
        18: "Subcutaneous_fat",
        19: "Visceral fat",
        20: "Muscles",
        21: "Bladder",
        22: "Filler"
    },
    "clin_pt_fdg_brain_v1": {
        0: "background",
        1: "R-Hippocampus",
        2: "L-Hippocampus",
        3: "R-Amygdala",
        4: "L-Amygdala",
        5: "R-Anterior-temporal-lobe-medial-part",
        6: "L-Anterior-temporal-lobe-medial-part",
        7: "R-Anterior-temporal-lobe-lateral-part",
        8: "L-Anterior-temporal-lobe-lateral-part",
        9: "R-Parahippocampal-and-ambient-gyri",
        10: "L-Parahippocampal-and-ambient-gyri",
        11: "R-Superior-temporal-gyrus-posterior-part",
        12: "L-Superior-temporal-gyrus-posterior-part",
        13: "R-Middle and inferior temporal gyrus",
        14: "L-Middle and inferior temporal gyrus",
        15: "R-Fusiform gyrus",
        16: "L-Fusiform gyrus",
        17: "R-Cerebellum",
        18: "L-Cerebellum",
        19: "Brainstem",
        20: "L-Insula",
        21: "R-Insula",
        22: "L-Lateral remainder of occipital lobe",
        23: "R-Lateral remainder of occipital lobe",
        24: "L-Cingulate gyrus gyrus cinguli anterior part",
        25: "R-Cingulate gyrus gyrus cinguli anterior part",
        26: "L-Cingulate gyrus gyrus cinguli posterior part",
        27: "R-Cingulate gyrus gyrus cinguli posterior part",
        28: "L-Middle frontal gyrus",
        29: "R-Middle frontal gyrus",
        30: "L-Posterior temporal lobe",
        31: "R-Posterior temporal lobe",
        32: "L-Inferiolateral remainder of parietal lobe",
        33: "R-Inferiolateral remainder of parietal lobe",
        34: "L-Caudate nucleus",
        35: "R-Caudate nucleus",
        36: "L-Nucleus accumbens",
        37: "R-Nucleus accumbens",
        38: "L-Putamen",
        39: "R-Putamen",
        40: "L-Thalamus",
        41: "R-Thalamus",
        42: "L-Pallidum",
        43: "R-Pallidum",
        44: "Corpus callosum",
        45: "R-Lateral ventricle excluding temporal horn",
        46: "L-Lateral ventricle excluding temporal horn",
        47: "R-Lateral ventricle, temporal horn",
        48: "L-Lateral ventricle, temporal horn",
        49: "Third ventricle",
        50: "L-Precentral gyrus",
        51: "R-Precentral gyrus",
        52: "L-Straight gyrus",
        53: "R-Straight gyrus",
        54: "L-Anterior orbital gyrus",
        55: "R-Anterior orbital gyrus",
        56: "L-Inferior frontal gyrus",
        57: "R-Inferior frontal gyrus",
        58: "L-Superior frontal gyrus",
        59: "R-Superior frontal gyrus",
        60: "L-Postcentral gyrus",
        61: "R-Postcentral gyrus",
        62: "L-Superior parietal gyrus",
        63: "R-Superior parietal gyrus",
        64: "L-Lingual gyrus",
        65: "R-Lingual gyrus",
        66: "L-Cuneus",
        67: "R-Cuneus",
        68: "L-Medial orbital gyrus",
        69: "R-Medial orbital gyrus",
        70: "L-Lateral orbital gyrus",
        71: "R-Lateral orbital gyrus",
        72: "L-Posterior orbital gyrus",
        73: "R-Posterior orbital gyrus",
        74: "L-Substantia nigra",
        75: "R-Substantia nigra",
        76: "L-Subgenual frontal cortex",
        77: "R-Subgenual frontal cortex",
        78: "L-Subcallosal area",
        79: "R-Subcallosal area",
        80: "L-Pre-subgenual frontal cortex",
        81: "R-Pre-subgenual frontal cortex",
        82: "L-Superior temporal gyrus anterior part",
        83: "R-Superior temporal gyrus anterior part"
    },
    "clin_ct_ALPACA":{
        1: "heart_ventricle_left",
        2: "heart_ventricle_right",
        3: "pulmonary_artery",
        4: "iliac_artery_left",
        5: "iliac_artery_right",
        6: "aorta"}
    # More index-to-name dictionaries for other models...
}
"""

This module contains the constants that are used in the moosez.

Constants are values that are fixed and do not change during the execution of a program. They are used to store values that are used repeatedly throughout the program, such as file paths, folder names, and display parameters.

This module contains the following constants:

- `NNUNET_RESULTS_FOLDER`: A constant that stores the path to the folder that contains the trained models for the nnUNet algorithm.
- `MOOSEZ_MODEL_FOLDER`: A constant that stores the path to the folder that contains the 3D full resolution model for the moosez algorithm.
- `ALLOWED_MODALITIES`: A constant that stores a list of allowed modalities for the moosez algorithm.
- `TEMP_FOLDER`: A constant that stores the name of the temporary folder used by the moosez algorithm.
- `ANSI_ORANGE`: A constant that stores the ANSI color code for orange.
- `ANSI_GREEN`: A constant that stores the ANSI color code for green.
- `ANSI_VIOLET`: A constant that stores the ANSI color code for violet.
- `ANSI_RESET`: A constant that stores the ANSI color code for resetting the color.
- `TRACER_FDG`: A constant that stores the name of the tracer used by the moosez algorithm.
- `SEGMENTATIONS_FOLDER`: A constant that stores the name of the folder that contains the segmentations generated by the moosez algorithm.
- `STATS_FOLDER`: A constant that stores the name of the folder that contains the statistics generated by the moosez algorithm.
- `MATRIX_THRESHOLD`: A constant that stores the matrix threshold used by the moosez algorithm.
- `Z_AXIS_THRESHOLD`: A constant that stores the z-axis threshold used by the moosez algorithm.
- `MARGIN_PADDING`: A constant that stores the margin padding used by the moosez algorithm.
- `INTERPOLATION`: A constant that stores the interpolation method used by the moosez algorithm.
- `CHUNK_THRESHOLD`: A constant that stores the chunk threshold used by the moosez algorithm.
- `TUMOR_LABEL`: A constant that stores the label used for tumors by the moosez algorithm.
- `RESAMPLED_IMAGE_FILE_NAME`: A constant that stores the name of the resampled image file used by the moosez algorithm.
- `RESAMPLED_MASK_FILE_NAME`: A constant that stores the name of the resampled mask file used by the moosez algorithm.
- `CHUNK_FILENAMES`: A constant that stores the names of the chunk files used by the moosez algorithm.
- `CHUNK_PREFIX`: A constant that stores the prefix used for chunk files by the moosez algorithm.
- `MIP_ROTATION_STEP`: A constant that stores the MIP rotation step used by the moosez algorithm.
- `DISPLAY_VOXEL_SPACING`: A constant that stores the display voxel spacing used by the moosez algorithm.
- `FRAME_DURATION`: A constant that stores the frame duration used by the moosez algorithm.
- `ORGAN_INDICES`: A constant that stores a dictionary of index-to-name mappings for various organs used by the moosez algorithm.

This module is imported by other modules in the moosez package and the constants are used throughout the package to provide fixed values that are used repeatedly.
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging

# ----------------------------------------------------------------------------------------------------------------------
# Author: Lalith Kumar Shiyam Sundar
# Institution: Medical University of Vienna
# Research Group: Quantitative Imaging and Medical Physics (QIMP) Team
# Date: 13.02.2023
# Version: 2.0.0
#
# Description:
# This module contains the urls and filenames of the models and binaries that are required for the moosez.
#
# Usage:
# The variables in this module can be imported and used in other modules within the moosez to download the necessary
# binaries and models for the moosez.
#
# ----------------------------------------------------------------------------------------------------------------------
import torch
from moosez import constants

# This list contains the unique identifiers for all available pre-trained models.
# To make your model available, add its unique identifier to this list.
# The model name or the unique identifier has a specific syntax:
# 'clin' or 'preclin' (indicating Clinical or Preclinical),
# modality tag (like 'ct', 'pt', 'mr'), and then the tissue of interest.

AVAILABLE_MODELS = ["clin_ct_lungs",
                    "clin_ct_organs",
                    "clin_ct_body",
                    "preclin_mr_all",
                    "clin_ct_ribs",
                    "clin_ct_muscles",
                    "clin_ct_peripheral_bones",
                    "clin_ct_fat",
                    "clin_ct_vertebrae",
                    "clin_ct_cardiac",
                    "clin_ct_digestive",
                    "preclin_ct_legs",
                    "clin_ct_all_bones_v1",
                    "clin_ct_PUMA",
                    "clin_pt_fdg_brain_v1",
                    "clin_ct_ALPACA"]

# This dictionary holds the pre-trained models available in MooseZ library.
# Each key is a unique model identifier following a specific syntax mentioned above
# It should have the same name mentioned in AVAILABLE_MODELS list.
# Each value is a dictionary containing the following keys:
#    - url: The URL where the model files can be downloaded.
#    - filename: The filename of the model's zip file.
#    - directory: The directory where the model files will be extracted.
#    - trainer: The type of trainer used to train the model.
#    - voxel_spacing: The voxel spacing used in the model in the form [x, y, z], this is basically the median voxel
#    spacing generated by nnunetv2, and you can find this in the plans.json file of the model.
#    - multilabel_prefix: A prefix to distinguish between different types of labels in multi-label models.
#
# To include your own model, add a new entry to this dictionary following the above format.

MODELS = {
    "clin_ct_lungs": {
        "url": "https://ucd-emic-muv.s3.us-west-2.amazonaws.com/moose/clin_ct_lungs_24062023.zip",
        "filename": "Dataset333_HMS3dlungs.zip",
        "directory": "Dataset333_HMS3dlungs",
        "trainer": "nnUNetTrainer_2000epochs_NoMirroring",
        "voxel_spacing": [1.5, 1.5, 1.5],
        "multilabel_prefix": "CT_Lungs_"
    },
    "clin_ct_organs": {
        "url": "https://ucd-emic-muv.s3.us-west-2.amazonaws.com/moose/MOOSEv2_bspline_organs23062023.zip",
        "filename": "Dataset123_Organs.zip",
        "directory": "Dataset123_Organs",
        "trainer": "nnUNetTrainer_2000epochs_NoMirroring",
        "voxel_spacing": [1.5, 1.5, 1.5],
        "multilabel_prefix": "CT_Organs_"
    },
    "preclin_mr_all": {
        "url": "https://ucd-emic-muv.s3.us-west-2.amazonaws.com/moose/preclin_mr_all_05122023.zip",
        "filename": "Dataset234_minimoose.zip",
        "directory": "Dataset234_minimoose_mr",
        "trainer": "nnUNetTrainer",
        "voxel_spacing": [0.4000000059604645, 0.4000000059604645, 0.4000000059604645],
        "multilabel_prefix": "Preclin_MR_all_"
    },
    "clin_ct_body": {
        "url": "https://ucd-emic-muv.s3.us-west-2.amazonaws.com/moose/clin_ct_body_27112023.zip",
        "filename": "Dataset001_body.zip",
        "directory": "Dataset001_body",
        "trainer": "nnUNetTrainer",
        "voxel_spacing": [5, 5, 5],
        "multilabel_prefix": "CT_Body_"
    },
    "clin_ct_ribs": {
        "url": "https://ucd-emic-muv.s3.us-west-2.amazonaws.com/moose/clin_ct_ribs_25082023.zip",
        "filename": "Dataset444_Ribs.zip",
        "directory": "Dataset444_Ribs",
        "trainer": "nnUNetTrainer_2000epochs_NoMirroring",
        "voxel_spacing": [1.5, 1.5, 1.5],
        "multilabel_prefix": "CT_Ribs_"
    },
    "clin_ct_muscles": {
        "url": "https://ucd-emic-muv.s3.us-west-2.amazonaws.com/moose/clin_ct_muscles_28082023.zip",
        "filename": "Dataset555_Muscles.zip",
        "directory": "Dataset555_Muscles",
        "trainer": "nnUNetTrainer_2000epochs_NoMirroring",
        "voxel_spacing": [1.5, 1.5, 1.5],
        "multilabel_prefix": "CT_Muscles_"
    },
    "clin_ct_peripheral_bones": {
        "url": "https://ucd-emic-muv.s3.us-west-2.amazonaws.com/moose/clin_ct_peripheral_bones_28082023.zip",
        "filename": "Dataset666_Peripheral-Bones.zip",
        "directory": "Dataset666_Peripheral-Bones",
        "trainer": "nnUNetTrainer_2000epochs_NoMirroring",
        "voxel_spacing": [1.5, 1.5, 1.5],
        "multilabel_prefix": "CT_Peripheral-Bones_"
    },
    "clin_ct_fat": {
        "url": "https://ucd-emic-muv.s3.us-west-2.amazonaws.com/moose/clin_ct_fat_31082023.zip",
        "filename": "Dataset777_Fat.zip",
        "directory": "Dataset777_Fat",
        "trainer": "nnUNetTrainer_2000epochs_NoMirroring",
        "voxel_spacing": [1.5, 1.5, 1.5],
        "multilabel_prefix": "CT_Fat_"
    },
    "clin_ct_vertebrae": {
        "url": "https://ucd-emic-muv.s3.us-west-2.amazonaws.com/moose/clin_ct_vertebrae_da5_03102023.zip",
        "filename": "Dataset111_Vertebrae.zip",
        "directory": "Dataset111_Vertebrae",
        "trainer": "nnUNetTrainer_2000_epochs_DA5NoMirroring",
        "voxel_spacing": [1.5, 1.5, 1.5],
        "multilabel_prefix": "CT_Vertebrae_"
    },
    "clin_ct_cardiac": {
        "url": "https://ucd-emic-muv.s3.us-west-2.amazonaws.com/moose/clin_ct_cardiac_10092023.zip",
        "filename": "Dataset888_Cardiac.zip",
        "directory": "Dataset888_Cardiac",
        "trainer": "nnUNetTrainer_2000epochs_NoMirroring",
        "voxel_spacing": [1.5, 1.5, 1.5],
        "multilabel_prefix": "CT_Cardiac_"
    },
    "clin_ct_digestive": {
        "url": "https://ucd-emic-muv.s3.us-west-2.amazonaws.com/moose/clin_ct_digestive10092023.zip",
        "filename": "Dataset999_Digestive.zip",
        "directory": "Dataset999_Digestive",
        "trainer": "nnUNetTrainer_2000epochs_NoMirroring",
        "voxel_spacing": [1.5, 1.5, 1.5],
        "multilabel_prefix": "CT_Digestive_"
    },
    "preclin_ct_legs": {
        "url": "https://ucd-emic-muv.s3.us-west-2.amazonaws.com/moose/preclin_ct_legs_05122023.zip",
        "filename": "Dataset256_Preclin_leg_muscles.zip",
        "directory": "Dataset256_Preclin_leg_muscles",
        "trainer": "nnUNetTrainerNoMirroring",
        "voxel_spacing": [0.18000000715255737, 0.18000000715255737, 0.18000000715255737],
        "multilabel_prefix": "Preclin_CT_legs_"
    },
    "clin_ct_all_bones_v1": {
        "url": "https://ucd-emic-muv.s3.us-west-2.amazonaws.com/moose/clin_ct_all_bones_25102023.zip",
        "filename": "Dataset600_Original_bones.zip",
        "directory": "Dataset600_Original_bones",
        "trainer": "nnUNetTrainer_2000epochs",
        "voxel_spacing": [1.5, 1.5, 1.5],
        "multilabel_prefix": "Clin_CT_all_bones_"
    },
    "clin_ct_PUMA": {
        "url": "https://ucd-emic-muv.s3.us-west-2.amazonaws.com/moose/clin_ct_PUMA_06112023.zip",
        "filename": "Dataset002_PUMA.zip",
        "directory": "Dataset002_PUMA",
        "trainer": "nnUNetTrainer_2000epochs",
        "voxel_spacing": [1.5, 1.5, 1.5],
        "multilabel_prefix": "Clin_CT_PUMA_"
    },
    "clin_pt_fdg_brain_v1": {
        "url": "https://ucd-emic-muv.s3.us-west-2.amazonaws.com/moose/clin_fdg_pt_brain_v1_17112023.zip",
        "filename": "Dataset100_Brain_v1.zip",
        "directory": "Dataset100_Brain_v1",
        "trainer": "nnUNetTrainer_2000epochs_NoMirroring",
        "voxel_spacing": [2.03125, 2.0862598419189453, 2.0862600803375244],
        "multilabel_prefix": "Clin_PT_FDG_brain_"
    },
    "clin_ct_ALPACA": {
        "url": "https://ucd-emic-muv.s3.us-west-2.amazonaws.com/moose/clin_ct_ALPACA.zip",
        "filename": "Dataset080_ALPACA.zip",
        "directory": "Dataset080_ALPACA",
        "trainer": "nnUNetTrainer_2000epochs_NoMirroring",
        "voxel_spacing": [1.5, 1.5, 1.5],
        "multilabel_prefix": "Clin_CT_ALPACA_"
    }
}


# This function returns a dictionary indicating the expected modality for a given model_name, the imaging technique,
# the type of tissue to be segmented. The model_name should be the same as the unique identifier mentioned in the
# MODELS dictionary above and the AVAILABLE_MODELS list.
# If the model_name is not found, it logs an error message and returns an error message.
#
# If you add your own model, update this function to return the expected modality dictionary for your model.

def expected_modality(model_name: str) -> dict:
    """
    Display expected modality for the model.
    :param model_name: The name of the model.
    :return: The expected modality for the model.
    """
    models = {
        "clin_ct_lungs": {"Imaging": "Clinical", "Modality": "CT", "Tissue of interest": "Lungs"},
        "clin_ct_organs": {"Imaging": "Clinical", "Modality": "CT", "Tissue of interest": "Organs"},
        "clin_ct_body": {"Imaging": "Clinical", "Modality": "CT", "Tissue of interest": "Body, Arms, legs, head"},
        "preclin_mr_all": {"Imaging": "Pre-clinical", "Modality": "MR", "Tissue of interest": "All regions"},
        "clin_ct_ribs": {"Imaging": "Clinical", "Modality": "CT", "Tissue of interest": "Ribs"},
        "clin_ct_muscles": {"Imaging": "Clinical", "Modality": "CT", "Tissue of interest": "Muscles"},
        "clin_ct_peripheral_bones": {"Imaging": "Clinical", "Modality": "CT", "Tissue of interest": "Peripheral Bones"},
        "clin_ct_fat": {"Imaging": "Clinical", "Modality": "CT", "Tissue of interest": "Fat"},
        "clin_ct_vertebrae": {"Imaging": "Clinical", "Modality": "CT", "Tissue of interest": "Vertebrae"},
        "clin_ct_cardiac": {"Imaging": "Clinical", "Modality": "CT", "Tissue of interest": "Cardiac"},
        "clin_ct_digestive": {"Imaging": "Clinical", "Modality": "CT", "Tissue of interest": "Digestive"},
        "preclin_ct_legs": {"Imaging": "Pre-clinical", "Modality": "CT", "Tissue of interest": "Legs"},
        "clin_ct_all_bones_v1": {"Imaging": "Clinical", "Modality": "CT", "Tissue of interest": "All bones"},
        "clin_ct_PUMA": {"Imaging": "Clinical", "Modality": "CT", "Tissue of interest": "PUMA tissues"},
        "clin_pt_fdg_brain_v1": {"Imaging": "Clinical", "Modality": "PT", "Tissue of interest": "Brain regions"},
        "clin_ct_ALPACA": {"Imaging": "Clinical", "Modality": "CT", "Tissue of interest": "ALPACA tissues"}
    }

    if model_name in models:
        model = models[model_name]
        model["Model name"] = model_name
        return model

    logging.error(" Requested model is not available. Please check the model name.")
    return {"Error": "Requested model is not available. Please check the model name."}


# This function maps the model name to the task number. This is the number that comes after Dataset in DatasetXXXX,
# after nnunetv2 training. If your model folder is Dataset123, then the task number is 123.
# It checks for known model names and returns the associated task number, this is ABSOLUTELY NEEDED for the moosez to
# work. If the provided model name doesn't match any known model, it raises an exception.

# When adding your own model, update this function to return the task number associated with your model.

def map_model_name_to_task_number(model_name: str):
    """
    Maps the model name to the task number.
    :param model_name: The name of the model.
    :return: The task number.
    """
    if model_name == "clin_ct_lungs":
        return '333'
    elif model_name == "clin_ct_organs":
        return '123'
    elif model_name == "preclin_mr_all":
        return '234'
    elif model_name == "clin_ct_body":
        return '001'
    elif model_name == "clin_ct_ribs":
        return '444'
    elif model_name == "clin_ct_muscles":
        return "555"
    elif model_name == "clin_ct_peripheral_bones":
        return "666"
    elif model_name == "clin_ct_fat":
        return "777"
    elif model_name == "clin_ct_vertebrae":
        return "111"
    elif model_name == "clin_ct_cardiac":
        return "888"
    elif model_name == "clin_ct_digestive":
        return "999"
    elif model_name == "preclin_ct_legs":
        return "256"
    elif model_name == "clin_ct_all_bones_v1":
        return "600"
    elif model_name == "clin_ct_PUMA":
        return "002"
    elif model_name == "clin_pt_fdg_brain_v1":
        return "100"
    elif model_name == "clin_ct_ALPACA":
        return "080"
    else:
        raise Exception(f"Error: The model name '{model_name}' is not valid.")


def check_device() -> str:
    """
    This function checks the available device for running predictions, considering CUDA and MPS (for Apple Silicon).

    Returns:
        str: The device to run predictions on, either "cpu", "cuda", or "mps".
    """
    # Check for CUDA
    if torch.cuda.is_available():
        device_count = torch.cuda.device_count()
        print(f" CUDA is available with {device_count} GPU(s). Predictions will be run on GPU.")
        return "cuda"
    # Check for MPS (Apple Silicon)
    elif torch.backends.mps.is_available():
        print(" Apple MPS backend is available. Predictions will be run on Apple Silicon GPU.")
        return "mps"
    elif not torch.backends.mps.is_built():
        print(" MPS not available because the current PyTorch install was not built with MPS enabled.")
        return "cpu"
    else:
        print(" CUDA/MPS not available. Predictions will be run on CPU.")
        return "cpu"
